-- modules/assets/import :: a simple assets importer, to import easily everything into
-- the asset manager.

--[[
  Copyright © 2019 Kazhnuz

  Permission is hereby granted, free of charge, to any person obtaining a copy of
  this software and associated documentation files (the "Software"), to deal in
  the Software without restriction, including without limitation the rights to
  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
  the Software, and to permit persons to whom the Software is furnished to do so,
  subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]

local Assets = Object:extend()

local cwd  = (...):gsub('%.init$', '') .. ".types."

local Texture     = require(cwd .. "texture")

local Sprite      = require(cwd .. "sprites")
local Font        = require(cwd .. "fonts")
local ImageFont   = require(cwd .. "imagefonts")

local Tileset     = require(cwd ..  "tileset")
local Autotile    = require(cwd ..  "autotile")
local Background  = require(cwd ..  "background")

local SFX         = require(cwd .. "sfx")

-- INIT FUNCTIONS
-- Initilizing and configuring option

function Assets:new(isGlobal)
    self.isGlobal = isGlobal
end

function Assets:clear()
    if (self.isGlobal) then
        core.assets:clear()
    else
        core.assets:clearLocal()
    end
end

-- WRAPPER FUNCTIONS
-- Basic wrappers to ensure compatibility

function Assets:get(key)
  return core.assets:get(key)
end

function Assets:getFont(key)
    return self:getWithType(key, "font")
end

function Assets:getWithType(key, type)
  return core.assets:get(key)
end

function Assets:draw(name, x, y, r, sx, sy, ox, oy, kx, ky)
    core.assets:draw(name, x, y, r, sx, sy, ox, oy, kx, ky)
end

function Assets:newSFX(name, filepath)
    self:addSFX(name, filepath)
end

function Assets:addImage(name, filename)
  self:addTexture(name, filename)
end

function Assets:playSFX(name)
    core.assets:playSFX(name)
end

function Assets:drawImage(name, x, y, r, sx, sy, ox, oy, kx, ky)
    self:draw(name, x, y, r, sx, sy, ox, oy, kx, ky)
end

-- Music

function Assets:setMusic(filename, loop)
    core.debug:logDebug("assets", "Assets:setMusic is deprecated")
    core.music:addMusic(filename, true)
end

function Assets:silence()
    core.debug:logDebug("assets", "Assets:silence is deprecated")
    core.music:silence()
end

function Assets:resetMusic()
    core.debug:logDebug("assets", "Assets:resetMusic is deprecated")
    core.music:purge()
end

function Assets:playMusic()
    core.debug:logDebug("assets", "Assets:playMusic is deprecated")
    core.music:playMusic()
end

-- Activity

function Assets:setActivity(activity)
    core.assets:setActivity(activity)
end

function Assets:switchActivity()
    core.assets:switchActivity()
end

function Assets:getActivity()
    core.assets:getActivity()
end

-- IMPORT FUNCTIONS
-- Easilly import assets

function Assets:batchImport(datafile)
    local datas = require(datafile)

    for asset_type, assets in pairs(datas) do
        if (asset_type == "autotiles") then
            self:importAutotiles(assets)
        elseif (asset_type == "backgrounds") then
            self:importBackgrounds(assets)
        elseif (asset_type == "fonts") then
            self:importFonts(assets)
        elseif (asset_type == "imagefonts") then
            self:importImageFonts(assets)
        elseif (asset_type == "images") then
            self:importTextures(assets)
        elseif (asset_type == "sprites") then
            self:importSprites(assets)
        elseif (asset_type == "textures") then
            self:importTextures(assets)
        elseif (asset_type == "tilesets") then
            self:importTilesets(assets)
        elseif (asset_type == "sfx") then
            self:importSFX(assets)
        else
            core.debug:logWarn("assets/importer", "unkown asset type " .. asset_type)
        end
    end
end

function Assets:importAutotiles(assets)
    for i, asset in ipairs(assets) do
        self:addAutotile(asset[1], asset[2])
    end
end

function Assets:importBackgrounds(assets)
    for i, asset in ipairs(assets) do
        self:addBackground(asset[1], asset[2])
    end
end

function Assets:importFonts(assets)
    for i, asset in ipairs(assets) do
        self:addFont(asset[1], asset[2], asset[3])
    end
end

function Assets:importImageFonts(assets)
    for i, asset in ipairs(assets) do
        self:addImageFont(asset[1], asset[2], asset[3])
    end
end

function Assets:importSprites(assets)
    for i, asset in ipairs(assets) do
        self:addSprite(asset[1], asset[2])
    end
end

function Assets:importTextures(assets)
    for i, asset in ipairs(assets) do
        self:addImage(asset[1], asset[2])
    end
end

function Assets:importTilesets(assets)
    for i, asset in ipairs(assets) do
        self:addTileset(asset[1], asset[2])
    end
end

function Assets:importSFX(assets)
    for i, asset in ipairs(assets) do
        self:addSFX(asset[1], asset[2])
    end
end

-- ADD FUNCTIONS
-- Different wrapper to create easily asset objects

function Assets:add(name, assetObject)
    core.assets:add(name, assetObject, self.isGlobal)
end

function Assets:addBackground(name, filepath)
    self:add(name, Background(filepath))
end

function Assets:addSprite(name, filepath)
    self:add(name, Sprite(filepath))
end

function Assets:addFont(key, filename, size)
    self:add(key, Font(filename, size))
end

function Assets:addImageFont(key, filename, extraspacing)
    self:add(key, ImageFont(filename, extraspacing))
end

function Assets:addTileset(name, filepath)
    self:add(name, Tileset(filepath))
end

function Assets:addAutotile(name, tilesize)
    self:add(name, Autotile(name, tilesize))
end

function Assets:addSFX(name, filepath)
    self:add(name, SFX(filepath))
end

function Assets:addTexture(name, filename)
  self:add(name, Texture(filename))
end

-- DEPRECATED FUNCTIONS
-- Don't do anything and will be removed soon

function Assets:clearFonts()
    core.debug:logDebug("assets", "Assets:clearFonts is deprecated")
end

function Assets:clearTileset()
    core.debug:logDebug("assets", "Assets:clearTileset is deprecated")
end

function Assets:clearAutotile()
    core.debug:logDebug("assets", "Assets:clearAutotile is deprecated")
end

function Assets:clearSFX()
    core.debug:logDebug("assets", "Assets:clearSFX is deprecated")
end

function Assets:clearBackgrounds()
    core.debug:logDebug("assets", "Assets:clearBackgrounds is deprecated")
end

function Assets:clearSprites()
    core.debug:logDebug("assets", "Assets:clearSprites is deprecated")
end

function Assets:clearImages()
    core.debug:logDebug("assets", "Assets:clearImages is deprecated")
end

return Assets
